import time
import numpy
import tensorflow as tf
import layers as L
import vat
import os

# FLAGS = tf.app.flags.FLAGS
tf.app.flags.DEFINE_string('data_dir', './dataset/cifar10/', 'where to store the dataset')
tf.app.flags.DEFINE_string('gpu', '0', "gpu")
tf.app.flags.DEFINE_string('device', '/gpu:0', "device")
tf.app.flags.DEFINE_string('dataset', 'cifar10', "{cifar10, svhn}")
tf.app.flags.DEFINE_string('log_dir_', "./log/cifar10/", "log_dir_")
tf.app.flags.DEFINE_integer('seed', 1, "initial random seed")
tf.app.flags.DEFINE_bool('validation', False, "")
tf.app.flags.DEFINE_integer('batch_size', 32, "the number of examples in a batch")
tf.app.flags.DEFINE_integer('ul_batch_size', 128, "the number of unlabeled examples in a batch")
tf.app.flags.DEFINE_integer('eval_batch_size', 100, "the number of eval examples in a batch")
tf.app.flags.DEFINE_integer('eval_freq', 5, "")
tf.app.flags.DEFINE_integer('num_epochs', 500, "the number of epochs for training")  # 500 * 400 = 20W
tf.app.flags.DEFINE_integer('epoch_decay_start', 460, "epoch of starting learning rate decay")
tf.app.flags.DEFINE_integer('num_iter_per_epoch', 400, "the number of updates per epoch")
tf.app.flags.DEFINE_float('learning_rate', 0.001, "initial leanring rate")
tf.app.flags.DEFINE_float('mom1', 0.9, "initial momentum rate")
tf.app.flags.DEFINE_float('mom2', 0.5, "momentum rate after epoch_decay_start")
tf.app.flags.DEFINE_string('method', 'peerent', "{supervised, peer, peerent}, peer: pixel-level in hidden space")
tf.app.flags.DEFINE_integer('n', 5, "neighbors: the number of image-level neighbors")
tf.app.flags.DEFINE_string('k', '10_20_20', "k_pixel: the number of pixel-level neighbors")
tf.app.flags.DEFINE_string('index_h', '0_1_2', "position of VAT, 0: input, 1, 2, 3")
tf.app.flags.DEFINE_string('alpha', '1.0', "coefficient of ")
tf.app.flags.DEFINE_string('patch', '3_1_1', "patch size in different positions")
tf.app.flags.DEFINE_bool('random_graph', False, '')
tf.app.flags.DEFINE_bool('aug_trans', False, "")
tf.app.flags.DEFINE_bool('aug_flip', False, "")

FLAGS = tf.app.flags.FLAGS
os.environ["CUDA_VISIBLE_DEVICES"] = FLAGS.gpu
if FLAGS.aug_trans or FLAGS.aug_flip:
    FLAGS.log_dir_ += 'Aug_'
if FLAGS.random_graph:
    FLAGS.log_dir_ += 'RandG_'
else:
    FLAGS.log_dir_ += 'KNNG_'
FLAGS.log_dir_ = FLAGS.log_dir_ + 'eps_' + str(
    FLAGS.epsilon) + '_h_' + FLAGS.index_h + '_alpha_' + FLAGS.alpha + '_patch_' + FLAGS.patch + '_n_' + str(
    FLAGS.n) + '_k_' + FLAGS.k + '_' + time.strftime('%Y_%m_%d_%H_%M_%S', time.localtime(time.time()))
INDEX_H = [int(i) for i in FLAGS.index_h.split('_')]
ALPHA = [float(i) for i in FLAGS.alpha.split('_')]
PSIZE = [int(i) for i in FLAGS.patch.split('_')]
KPIX = [int(i) for i in FLAGS.k.split('_')]
if (len(KPIX) == 1) and (len(INDEX_H) > 1):
    KPIX = KPIX * len(INDEX_H)
if (len(ALPHA) == 1) and (len(INDEX_H) > 1):
    ALPHA = ALPHA * len(INDEX_H)
if (len(PSIZE) == 1) and (len(INDEX_H) > 1):
    PSIZE = PSIZE * len(INDEX_H)
if FLAGS.dataset == 'cifar10':
    from cifar10 import inputs, unlabeled_inputs
elif FLAGS.dataset == 'svhn':
    from svhn import inputs, unlabeled_inputs
else:
    raise NotImplementedError

config = tf.ConfigProto(allow_soft_placement=True)
# sess = tf.Session(config=config)

NUM_EVAL_EXAMPLES = 5000


def build_training_graph(x, y, ul_x, lr, mom):
    global_step = tf.get_variable(
        name="global_step",
        shape=[],
        dtype=tf.float32,
        initializer=tf.constant_initializer(0.0),
        trainable=False,
    )
    # (1) compute supervised loss
    logit, _ = vat.forward(x)
    nll_loss = L.ce_loss(logit, y)  # cross entropy loss
    # (2) compute VAT loss based on unlabeled data
    with tf.variable_scope(tf.get_variable_scope(), reuse=True):
        if FLAGS.method == 'peer':
            ul_logit, ul_feature = vat.forward(ul_x, is_training=True, update_batch_stats=False)
            vat_loss_graph = vat.virtual_adversarial_loss_graph_h(ul_x, ul_logit, ul_feature, FLAGS.random_graph,
                                                                  FLAGS.n, KPIX, INDEX_H, PSIZE, ALPHA)
            additional_loss = vat_loss_graph
        elif FLAGS.method == 'peerent':
            ul_logit, ul_feature = vat.forward(ul_x, is_training=True, update_batch_stats=False)
            # print('idxs:',idxs)
            vat_loss_graph = vat.virtual_adversarial_loss_graph_h(ul_x, ul_logit, ul_feature, FLAGS.random_graph,
                                                                  FLAGS.n, KPIX, INDEX_H, PSIZE, ALPHA)
            ent_loss = L.entropy_y_x(ul_logit)
            additional_loss = vat_loss_graph + ent_loss
        elif FLAGS.method == 'supervised':
            additional_loss = 0
        else:
            print('Method name is erroneous!')
            raise NotImplementedError
        loss = nll_loss + additional_loss

    opt = tf.train.AdamOptimizer(learning_rate=lr, beta1=mom)
    tvars = tf.trainable_variables()
    grads_and_vars = opt.compute_gradients(loss, tvars)  # loss to parameters to get gradients = loss.backward()
    train_op = opt.apply_gradients(grads_and_vars, global_step=global_step)  # updating parameters via optimizer
    return loss, train_op, global_step


def build_eval_graph(x, y, ul_x):  # X - CNN - loss
    losses = {}
    logit, _ = vat.forward(x, is_training=False, update_batch_stats=False)
    nll_loss = L.ce_loss(logit, y)
    losses['NLL'] = nll_loss
    acc = L.accuracy(logit, y)
    losses['Acc'] = acc
    scope = tf.get_variable_scope()
    scope.reuse_variables()
    # at_loss = vat.adversarial_loss(x, y, nll_loss, is_training=False)
    # losses['AT_loss'] = at_loss
    ul_logit, ul_feature = vat.forward(ul_x, is_training=False, update_batch_stats=False)
    # vat_loss = vat.virtual_adversarial_loss(ul_x, ul_logit, is_training=False)
    # vat_loss=vat.virtual_adversarial_loss_graph_h(ul_x, ul_logit, last_logits, FLAGS.n, FLAGS.k, INDEX_H, PSIZE, ALPHA, is_training=False)
    vat_loss = vat.virtual_adversarial_loss_graph_h(ul_x, ul_logit, ul_feature, FLAGS.random_graph, FLAGS.n, KPIX,
                                                    INDEX_H, PSIZE, ALPHA, is_training=False)
    losses['VAT_loss'] = vat_loss
    return losses


def main(_):
    print('Epsilon: ', FLAGS.epsilon, FLAGS.top_bn)
    numpy.random.seed(seed=FLAGS.seed)
    tf.set_random_seed(numpy.random.randint(1234))
    with tf.Graph().as_default() as g:
        # (1) load data: train, eval
        with tf.device("/cpu:0"):
            images, labels = inputs(batch_size=FLAGS.batch_size, train=True, validation=FLAGS.validation, shuffle=True)
            ul_images = unlabeled_inputs(batch_size=FLAGS.ul_batch_size, validation=FLAGS.validation, shuffle=True)
            images_eval_train, labels_eval_train = inputs(batch_size=FLAGS.eval_batch_size, train=True,
                                                          validation=FLAGS.validation, shuffle=True)
            ul_images_eval_train = unlabeled_inputs(batch_size=FLAGS.eval_batch_size, validation=FLAGS.validation,
                                                    shuffle=True)
            images_eval_test, labels_eval_test = inputs(batch_size=FLAGS.eval_batch_size, train=False,
                                                        validation=FLAGS.validation, shuffle=True)

        # (2) build computational graph
        with tf.device(FLAGS.device):
            lr = tf.placeholder(tf.float32, shape=[], name="learning_rate")
            mom = tf.placeholder(tf.float32, shape=[], name="momentum")
            with tf.variable_scope("CNN") as scope:
                # Build training graph
                loss, train_op, global_step = build_training_graph(images, labels, ul_images, lr, mom)
                scope.reuse_variables()  # train the same parameters
                # Build eval graph
                losses_eval_train = build_eval_graph(images_eval_train, labels_eval_train, ul_images_eval_train)
                losses_eval_test = build_eval_graph(images_eval_test, labels_eval_test, images_eval_test)

            init_op = tf.global_variables_initializer()
        # (3) define logging
        if not FLAGS.log_dir_:
            logdir = None
            writer_train = None
            writer_test = None
        else:
            logdir = FLAGS.log_dir_
            writer_train = tf.summary.FileWriter(FLAGS.log_dir_ + "/train", g)
            writer_test = tf.summary.FileWriter(FLAGS.log_dir_ + "/test", g)

        saver = tf.train.Saver(tf.global_variables())
        sv = tf.train.Supervisor(
            is_chief=True,
            logdir=logdir,
            init_op=init_op,
            init_feed_dict={lr: FLAGS.learning_rate, mom: FLAGS.mom1},
            saver=saver,
            global_step=global_step,
            summary_op=None,
            summary_writer=None,
            save_model_secs=150, recovery_wait_secs=0)
        # (4) start training
        print("Training...")
        with sv.managed_session(config=config) as sess:
            for ep in range(FLAGS.num_epochs):
                if sv.should_stop():
                    break
                # (4.1) learning rate decay
                if ep < FLAGS.epoch_decay_start:
                    feed_dict = {lr: FLAGS.learning_rate, mom: FLAGS.mom1}
                else:
                    decayed_lr = ((FLAGS.num_epochs - ep) / float(
                        FLAGS.num_epochs - FLAGS.epoch_decay_start)) * FLAGS.learning_rate
                    feed_dict = {lr: decayed_lr, mom: FLAGS.mom2}
                # (4.2) 400 iteration each epoch
                sum_loss = 0
                start = time.time()
                pert_norm = []
                for i in range(FLAGS.num_iter_per_epoch):
                    # _, batch_loss, _ = sess.run([train_op, loss, global_step], feed_dict=feed_dict) # run loss, optimizer to update
                    sess_run_list = sess.run([train_op, loss, global_step] + tf.get_collection('pert_norm'),
                                             feed_dict=feed_dict)
                    batch_loss = sess_run_list[1]
                    pert_norm.append(sess_run_list[3:])
                    sum_loss += batch_loss
                pert_norm = numpy.array(pert_norm)
                pert_norm = numpy.mean(pert_norm, axis=0)
                end = time.time()
                print("Epoch:", ep, "CE_loss_train:", sum_loss / FLAGS.num_iter_per_epoch, 'PertNorm:', pert_norm,
                      "elapsed_time:{:.2f}".format(end - start), time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()))
                # (4.3) evaluation every 5 epoch
                if (ep + 1) % FLAGS.eval_freq == 0 or ep + 1 == FLAGS.num_epochs:
                    # Eval on training data
                    act_values_dict = {}
                    for key, _ in losses_eval_train.items():
                        act_values_dict[key] = 0
                    n_iter_per_epoch = NUM_EVAL_EXAMPLES / FLAGS.eval_batch_size
                    for i in range(int(n_iter_per_epoch)):
                        values = losses_eval_train.values()
                        act_values = sess.run(list(values))
                        for key, value in zip(act_values_dict.keys(), act_values):
                            act_values_dict[key] += value
                    summary = tf.Summary()
                    current_global_step = sess.run(global_step)
                    l1 = []
                    for key, value in act_values_dict.items():
                        l1.append('train-{}, {:.4f} '.format(key, value / n_iter_per_epoch))
                        summary.value.add(tag=key, simple_value=value / n_iter_per_epoch)
                    if writer_train is not None:
                        writer_train.add_summary(summary, current_global_step)

                    # Eval on test data
                    act_values_dict = {}
                    for key, _ in losses_eval_test.items():
                        act_values_dict[key] = 0
                    n_iter_per_epoch = NUM_EVAL_EXAMPLES / FLAGS.eval_batch_size
                    for i in range(int(n_iter_per_epoch)):
                        values = losses_eval_test.values()
                        act_values = sess.run(list(values))
                        for key, value in zip(act_values_dict.keys(), act_values):
                            act_values_dict[key] += value
                    summary = tf.Summary()
                    current_global_step = sess.run(global_step)
                    l2 = []
                    for key, value in act_values_dict.items():
                        l2.append('test-{}, {:.4f}'.format(key, value / n_iter_per_epoch))
                        summary.value.add(tag=key, simple_value=value / n_iter_per_epoch)
                    print(';'.join(l1), ';'.join(l2))
                    if writer_test is not None:
                        writer_test.add_summary(summary, current_global_step)

            saver.save(sess, sv.save_path, global_step=global_step)
        sv.stop()


if __name__ == "__main__":
    tf.app.run()
